#include <sys/types.h>
#include <sys/param.h>
#include <rpc/rpc.h>
#include <fcntl.h>
#include <unistd.h>
#include "rread.h"

#define MAX_GROUPS 50

read_results *
read_1(request, rq)
read_request *request;
struct svc_req *rq;
{
    static read_results read_results; /* Holds results of */
                                      /* the RPC          */
    struct authdes_cred *descp;   /* Authentication info  */
    uid_t uid;                    /* User ID of caller    */
    uid_t uid_orig;               /* User ID to restore   */
    gid_t gid;                    /* Group ID of caller   */
    gid_t gid_orig;               /* Group ID to restore  */
    int   gid_len;                /* Number of groups     */
    gid_t gidlist[MAX_GROUPS];    /* Auxiliary groups     */
    int n;                        /* Number of bytes read */
    int fd;                       /* fd of file to read   */
    char buf[MAX_DATA];           /* buffer to hold data  */

    switch (rq->rq_cred.oa_flavor) {
       case AUTH_DES:
          descp = (struct authdes_cred *)rq->rq_clntcred;
          if (!netname2user(descp->adc_fullname.name, &uid,
                  &gid, &gid_len, gidlist)) {
              read_results.status = 1;
              return(&read_results);
          }
          if (uid <= 100) {
              uid = UID_NOBODY;
              gid = GID_NOBODY;
              gid_len = 0;
          }
          break;
       default:
          uid = UID_NOBODY;
          gid = GID_NOBODY;
          gid_len = 0;
          break;
    }


    /*
     *  If the client requested too much data, return with
     *  an error indication
     */

    if (request->how_many > MAX_DATA) {
        read_results.status = 1;
        return(&read_results);
    }

    /*
     *  Set the effective user ID, effective group ID, and
     *  group list of the user calling this routine.
     */

    uid_orig = geteuid();
    gid_orig = getegid();

    if (  (setegid(gid) < 0)
       || (setgroups(gid_len, gidlist) < 0)
       || (seteuid(uid) < 0)) {
        read_results.status = 1;
        return(&read_results);
    }

    /*
     *  Now that the effective user ID and the effective
     *  group ID are set, perform the read operation...
     */

    read_results.status = 0;

    if (((fd = open(request->filename, O_RDONLY)) < 0)
      || (lseek(fd, (off_t)(request->start_location),
                                           SEEK_SET) < 0)
      || ((n = read(fd, buf, request->how_many)) < 0)) {
          read_results.status = 1;
    }

    /*
     *  Set your effective user and group ID back to their
     *  Original value
     */

    (void) setegid(gid_orig);
    (void) seteuid(uid_orig);

    /*
     *  Close the file, fill in the information and return.
     */

    if (fd >= 0)
        (void)close(fd);

    if (read_results.status == 1) {
        return(&read_results);
    }

    read_results.read_results_u.filedata.filedata_len = n;
    read_results.read_results_u.filedata.filedata_val = buf;

    return(&read_results);
}
